"""This module contains a single class: ConnectionFrame."""

import os
import tkinter
from tkinter import ttk, messagebox

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util


# pylint: disable-next=too-many-ancestors
class ConnectionFrame(ttk.Frame):
    """A ttk.Frame object that contains two ttk.Entry and
    two ttk.Button used to enter a connection string and
    encryption key and to connect to the database.
    """
    def __init__(self, parent: tkinter.Widget):
        super().__init__(parent)

        frame = ttk.Frame(self)
        frame.columnconfigure(1, weight=1)
        frame.pack(fill='both')

        ttk.Label(frame, text="Connection string:").grid(row=0, column=0, sticky='w')

        self.conn_entry = ttk.Entry(frame)
        self.conn_entry.grid(row=0, column=1, sticky='ew')

        self.conn_button = ttk.Button(frame, text="Connect", command=self._connect)
        self.conn_button.grid(row=0, column=2, sticky='e')

        ttk.Label(frame, text="Encryption key:").grid(row=1, column=0, sticky='w')

        self.key_entry = ttk.Entry(frame)
        self.key_entry.grid(row=1, column=1, sticky='ew')

        self.disconn_button = ttk.Button(frame, text="Disconnect", command=self._disconnect, state='disabled')
        self.disconn_button.grid(row=1, column=2, sticky='e')

        self._initial_connect()

    def _connect(self) -> None:
        """Validate the connection string and encryption key
        and connect to the database.
        """
        conn_string = self.conn_entry.get()
        crypto_key = self.key_entry.get()

        if not crypto_util.validate_key(crypto_key):
            messagebox.showerror("Invalid encryption key", "The entered encryption key is not a valid AES key.")
            return

        if db_util.connect(conn_string):
            crypto_util.set_key(crypto_key)
            self._set_state(True)

    def _disconnect(self) -> None:
        db_util.disconnect()
        crypto_util.set_key(None)
        self._set_state(False)

    def _set_state(self, connected: bool) -> None:
        if connected:
            self.conn_entry.configure(state='disabled')
            self.key_entry.configure(state='disabled')
            self.conn_button.configure(state='disabled')
            self.disconn_button.configure(state='normal')
        else:
            self.conn_entry.configure(state='normal')
            self.key_entry.configure(state='normal')
            self.conn_button.configure(state='normal')
            self.disconn_button.configure(state='disabled')

    def _initial_connect(self) -> None:
        """Check the environment for a connection string
        and encryption key and connect to the database if both
        are found.
        """
        conn_string = os.environ.get('OpenOrchestratorConnString', None)
        if conn_string:
            self.conn_entry.insert(0, conn_string)

        crypto_key = os.environ.get('OpenOrchestratorKey', None)
        if crypto_key:
            self.key_entry.insert(0, crypto_key)

        if conn_string and crypto_key:
            self._connect()

    def new_key(self):
        """Creates a new encryption key and inserts it
        into the key entry.
        """
        key = crypto_util.generate_key().decode()
        self.key_entry.delete(0, 'end')
        self.key_entry.insert(0, key)



"""This module defines ORM classes for queue elements."""

from datetime import datetime
import enum
from typing import Optional
import uuid

from sqlalchemy import String, Engine
from sqlalchemy.orm import Mapped, DeclarativeBase, mapped_column

from OpenOrchestrator.common import datetime_util

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class QueueStatus(enum.Enum):
    """An enum representing the status of a queue element."""
    NEW = 'New'
    IN_PROGRESS = 'In Progress'
    DONE = 'Done'
    FAILED = 'Failed'
    ABANDONED = 'Abandoned'


class Base(DeclarativeBase):
    """SqlAlchemy base class for all ORM classes in this module."""


class QueueElement(Base):
    """A class representing a queue element in the ORM."""
    __tablename__ = "Queues"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    queue_name: Mapped[str] = mapped_column(String(100), index=True)
    status: Mapped[QueueStatus] = mapped_column(default=QueueStatus.NEW)
    data: Mapped[Optional[str]] = mapped_column(String(2000))
    reference: Mapped[Optional[str]] = mapped_column(String(100))
    created_date: Mapped[datetime] = mapped_column(default=datetime.now, index=True)
    start_date: Mapped[Optional[datetime]]
    end_date: Mapped[Optional[datetime]]
    message: Mapped[Optional[str]] = mapped_column(String(1000))
    created_by: Mapped[Optional[str]] = mapped_column(String(100))

    def to_row_dict(self) -> dict:
        """Convert the object to a dict for display in a table."""
        return {
            "ID": self.id,
            "Reference": self.reference,
            "Status": self.status,
            "Data": self.data,
            "Created Date": datetime_util.format_datetime(self.created_date),
            "Start Date": datetime_util.format_datetime(self.start_date),
            "End Date": datetime_util.format_datetime(self.end_date),
            "Message": self.message,
            "Created By": self.created_by
        }


def create_tables(engine: Engine):
    """Create all SQL tables related to ORM classes in this module.

    Args:
        engine: The SqlAlchemy connection engine used to create the tables.
    """
    Base.metadata.create_all(engine)



"""This module is responsible for the layout and functionality of the Queues tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.queues import QueueStatus
from OpenOrchestrator.orchestrator.datetime_input import DatetimeInput


QUEUE_COLUMNS = [
    {'name': "Queue Name", 'label': "Queue Name", 'field': "Queue Name", 'align': 'left', 'sortable': True},
    {'name': "New", 'label': "New", 'field': "New", 'align': 'left', 'sortable': True},
    {'name': "In Progress", 'label': "In Progress", 'field': "In Progress", 'align': 'left', 'sortable': True},
    {'name': "Done", 'label': "Done", 'field': "Done", 'align': 'left', 'sortable': True},
    {'name': "Failed", 'label': "Failed", 'field': "Failed", 'align': 'left', 'sortable': True},
    {'name': "Abandoned", 'label': "Abandoned", 'field': "Abandoned", 'align': 'left', 'sortable': True}
]

ELEMENT_COLUMNS = [
    {'name': "Reference", 'label': "Reference", 'field': "Reference", 'align': 'left', 'sortable': True},
    {'name': "Status", 'label': "Status", 'field': "Status", 'align': 'left', 'sortable': True},
    {'name': "Data", 'label': "Data", 'field': "Data", 'align': 'left', 'sortable': True},
    {'name': "Message", 'label': "Message", 'field': "Message", 'align': 'left', 'sortable': True},
    {'name': "Created Date", 'label': "Created Date", 'field': "Created Date", 'align': 'left', 'sortable': True},
    {'name': "Start Date", 'label': "Start Date", 'field': "Start Date", 'align': 'left', 'sortable': True},
    {'name': "End Date", 'label': "End Date", 'field': "End Date", 'align': 'left', 'sortable': True},
    {'name': "Created By", 'label': "Created By", 'field': "Created By", 'align': 'left', 'sortable': True},
    {'name': "ID", 'label': "ID", 'field': "ID", 'align': 'left', 'sortable': True}
]


# pylint: disable-next=too-few-public-methods
class QueueTab():
    """The 'Queues' tab object. It contains tables and buttons for dealing with queues."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            self.queue_table = ui.table(title="Queues", columns=QUEUE_COLUMNS, rows=[], row_key='Queue Name', pagination={'rowsPerPage': 50, 'sortBy': 'Queue Name'}).classes("w-full")
            self.queue_table.on("rowClick", self._row_click)

    def update(self):
        """Update the queue table with data from the database."""
        queue_count = db_util.get_queue_count()

        # Convert queue count to row elements
        rows = []
        for queue_name, count in queue_count.items():
            row = {
                "Queue Name": queue_name,
                "New": count.get(QueueStatus.NEW, 0),
                "In Progress": count.get(QueueStatus.IN_PROGRESS, 0),
                "Done": count.get(QueueStatus.DONE, 0),
                "Failed": count.get(QueueStatus.FAILED, 0),
                "Abandoned": count.get(QueueStatus.ABANDONED, 0),
            }
            rows.append(row)

        self.queue_table.update_rows(rows)

    def _row_click(self, event):
        row = event.args[1]
        queue_name = row["Queue Name"]
        QueuePopup(queue_name)


# pylint: disable-next=too-few-public-methods
class QueuePopup():
    """A popup that displays queue elements in a queue."""
    def __init__(self, queue_name) -> None:
        self.queue_name = queue_name

        with ui.dialog(value=True).props('full-width full-height') as dialog, ui.card():
            with ui.row().classes("w-full"):
                self.from_input = DatetimeInput("From Date", on_change=self._update, allow_empty=True).style('margin-left: 1rem')
                self.to_input = DatetimeInput("To Date", on_change=self._update, allow_empty=True)

                self.limit_select = ui.select(
                    options=[100, 200, 500, 1000, "All"],
                    label="Limit",
                    value=100,
                    on_change=self._update).classes("w-24")

                ui.space()

                ui.switch("Dense", on_change=lambda e: self._dense_table(e.value))
                self._create_column_filter()
                ui.button(icon='refresh', on_click=self._update)
                ui.button(icon="close", on_click=dialog.close)
            with ui.scroll_area().classes("h-full"):
                self.table = ui.table(columns=ELEMENT_COLUMNS, rows=[], row_key='ID', title=queue_name, pagination=100).classes("w-full")

        self._update()

    def _dense_table(self, value: bool):
        """Change if the table is dense or not."""
        if value:
            self.table.props("dense")
        else:
            self.table.props(remove="dense")

    def _create_column_filter(self):
        """Create a menu with switches for toggling columns on and off."""
        def toggle(column: dict, visible: bool) -> None:
            column['classes'] = '' if visible else 'hidden'
            column['headerClasses'] = '' if visible else 'hidden'
            self.table.update()

        with ui.button("Columns", icon="menu"):
            with ui.menu(), ui.column().classes('gap-0 p-2'):
                for column in ELEMENT_COLUMNS:
                    ui.switch(column['label'], value=True, on_change=lambda e, column=column: toggle(column, e.value))

    def _update(self):
        """Update the table with values from the database."""
        limit = self.limit_select.value
        if limit == 'All':
            limit = 1_000_000_000

        from_date = self.from_input.get_datetime()
        to_date = self.to_input.get_datetime()

        queue_elements = db_util.get_queue_elements(self.queue_name, limit=limit, from_date=from_date, to_date=to_date)
        rows = [element.to_row_dict() for element in queue_elements]
        self.table.update_rows(rows)



"""This module defines ORM classes for triggers."""

from datetime import datetime
import enum
from typing import Optional
import uuid

from sqlalchemy import String, ForeignKey, Engine
from sqlalchemy.orm import Mapped, DeclarativeBase, mapped_column

from OpenOrchestrator.common import datetime_util

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class TriggerStatus(enum.Enum):
    """An enum representing the status of trigger processes."""
    IDLE = "Idle"
    RUNNING = "Running"
    FAILED = "Failed"
    DONE = "Done"
    PAUSED = "Paused"
    PAUSING = "Pausing"


class TriggerType(enum.Enum):
    """An enum representing the type of triggers."""
    SINGLE = "Single"
    SCHEDULED = "Scheduled"
    QUEUE = "Queue"


class Base(DeclarativeBase):
    """SqlAlchemy base class for all ORM classes in this module."""


class Trigger(Base):
    """A base class for all triggers in the ORM."""
    __tablename__ = "Triggers"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    trigger_name: Mapped[str] = mapped_column(String(100))
    process_name: Mapped[str] = mapped_column(String(100))
    last_run: Mapped[Optional[datetime]]
    process_path: Mapped[str] = mapped_column(String(250))
    process_args: Mapped[Optional[str]] = mapped_column(String(1000))
    process_status: Mapped[TriggerStatus] = mapped_column(default=TriggerStatus.IDLE)
    is_git_repo: Mapped[bool]
    is_blocking: Mapped[bool]
    type: Mapped[TriggerType]

    __mapper_args__ = {
        "polymorphic_on": "type",
        "polymorphic_abstract": True
    }

    def __repr__(self) -> str:
        return f"{self.trigger_name}: {self.type.value}"

    def to_row_dict(self) -> dict[str, str]:
        """Convert trigger to a row dictionary for display in a table."""
        return {
            "Trigger Name": self.trigger_name,
            "Type": self.type.value,
            "Status": self.process_status.value,
            "Process Name": self.process_name,
            "Last Run": datetime_util.format_datetime(self.last_run, "Never"),
            "ID": str(self.id)
        }


class SingleTrigger(Trigger):
    """A class representing single trigger objects in the ORM."""
    __tablename__ = "Single_Triggers"

    id: Mapped[uuid.UUID] = mapped_column(ForeignKey("Triggers.id"), primary_key=True)
    next_run: Mapped[datetime]

    __mapper_args__ = {"polymorphic_identity": TriggerType.SINGLE}

    def to_row_dict(self) -> dict[str, str]:
        row_dict = super().to_row_dict()
        row_dict["Next Run"] = datetime_util.format_datetime(self.next_run)
        return row_dict


class ScheduledTrigger(Trigger):
    """A class representing scheduled trigger objects in the ORM."""
    __tablename__ = "Scheduled_Triggers"

    id: Mapped[uuid.UUID] = mapped_column(ForeignKey("Triggers.id"), primary_key=True)
    cron_expr: Mapped[str] = mapped_column(String(200))
    next_run: Mapped[datetime]

    __mapper_args__ = {"polymorphic_identity": TriggerType.SCHEDULED}

    def to_row_dict(self) -> dict[str, str]:
        row_dict = super().to_row_dict()
        row_dict["Next Run"] = datetime_util.format_datetime(self.next_run)
        return row_dict


class QueueTrigger(Trigger):
    """A class representing queue trigger objects in the ORM."""
    __tablename__ = "Queue_Triggers"

    id: Mapped[uuid.UUID] = mapped_column(ForeignKey("Triggers.id"), primary_key=True)
    queue_name: Mapped[str] = mapped_column(String(200))
    min_batch_size: Mapped[int]

    __mapper_args__ = {"polymorphic_identity": TriggerType.QUEUE}

    def to_row_dict(self) -> dict[str, str]:
        row_dict = super().to_row_dict()
        row_dict["Next Run"] = "N/A"
        return row_dict


def create_tables(engine: Engine):
    """Create all SQL tables related to ORM classes in this module.

    Args:
        engine: The SqlAlchemy connection engine used to create the tables.
    """
    Base.metadata.create_all(engine)



"""This module is the entry point for the Orchestrator app. It contains a single class
that when created starts the application."""

import socket

from nicegui import ui, app

from OpenOrchestrator.orchestrator.tabs.trigger_tab import TriggerTab
from OpenOrchestrator.orchestrator.tabs.settings_tab import SettingsTab
from OpenOrchestrator.orchestrator.tabs.logging_tab import LoggingTab
from OpenOrchestrator.orchestrator.tabs.constants_tab import ConstantTab
from OpenOrchestrator.orchestrator.tabs.queue_tab import QueueTab


class Application():
    """The main application of Orchestrator.
    It contains a header and the four tabs of the application.
    """
    def __init__(self) -> None:
        with ui.header():
            with ui.tabs() as self.tabs:
                ui.tab('Triggers')
                ui.tab('Logs')
                ui.tab('Constants')
                ui.tab('Queues')
                ui.tab('Settings')

            ui.space()
            ui.button(icon="contrast", on_click=ui.dark_mode().toggle)
            ui.button(icon='refresh', on_click=self.update_tab)

        with ui.tab_panels(self.tabs, value='Settings', on_change=self.update_tab).classes('w-full') as self.tab_panels:
            self.t_tab = TriggerTab('Triggers')
            self.l_tab = LoggingTab("Logs")
            self.c_tab = ConstantTab("Constants")
            self.q_tab = QueueTab("Queues")
            SettingsTab('Settings')

        self._define_on_close()

        app.on_connect(self.update_loop)
        app.on_disconnect(app.shutdown)
        app.on_exception(lambda exc: ui.notify(exc, type='negative'))
        ui.run(title="Orchestrator", favicon='🤖', native=False, port=get_free_port(), reload=False)

    def update_tab(self):
        """Update the date in the currently selected tab."""
        match self.tab_panels.value:
            case 'Triggers':
                self.t_tab.update()
            case 'Logs':
                self.l_tab.update()
            case 'Constants':
                self.c_tab.update()
            case 'Queues':
                self.q_tab.update()

    async def update_loop(self):
        """Update the selected tab on a timer but only if the page is in focus."""
        try:
            in_focus = await ui.run_javascript("document.hasFocus()")
            if in_focus:
                self.update_tab()
        except TimeoutError:
            pass

        ui.timer(10, self.update_loop, once=True)

    def _define_on_close(self) -> None:
        """Tell the browser to ask for confirmation before leaving the page."""
        ui.add_body_html('''
            <script>
                window.addEventListener("beforeunload", (event) => event.preventDefault());
            </script>
            ''')


def get_free_port():
    """Get a free port by creating a new socket and bind it
    on port 0 allowing the os to select the port.
    https://docs.python.org/3/library/socket.html#socket.create_connection

    Returns:
        A port number that should be free to use.
    """
    with socket.socket() as sock:
        sock.bind(("", 0))
        port = sock.getsockname()[1]

    return port


if __name__ in {'__main__', '__mp_main__'}:
    Application()



"""This module is responsible for the layout and functionality of the Logging tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.logs import LogLevel
from OpenOrchestrator.orchestrator.datetime_input import DatetimeInput


COLUMNS = [
    {'name': "Log Time", 'label': "Log Time", 'field': "Log Time", 'align': 'left', 'sortable': True},
    {'name': "Process Name", 'label': "Process Name", 'field': "Process Name", 'align': 'left'},
    {'name': "Level", 'label': "Level", 'field': "Level", 'align': 'left'},
    {'name': "Message", 'label': "Message", 'field': "Message", 'align': 'left', ':format': 'value => value.length < 100 ? value : value.substring(0, 100)+"..."'},
    {'name': "ID", 'label': "ID", 'field': "ID", 'headerClasses': 'hidden', 'classes': 'hidden'}
]


# pylint: disable-next=too-few-public-methods
class LoggingTab():
    """The 'Logs' tab object."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            with ui.row():
                self.from_input = DatetimeInput("From Date", on_change=self.update, allow_empty=True)
                self.to_input = DatetimeInput("To Date", on_change=self.update, allow_empty=True)
                self.process_input = ui.select(["All"], label="Process Name", value="All", on_change=self.update).classes("w-48")
                self.level_input = ui.select(["All", "Trace", "Info", "Error"], value="All", label="Level", on_change=self.update).classes("w-48")
                self.limit_input = ui.select([100, 200, 500, 1000], value=100, label="Limit", on_change=self.update).classes("w-24")

            self.logs_table = ui.table(title="Logs", columns=COLUMNS, rows=[], row_key='ID', pagination=50).classes("w-full")
            self.logs_table.on("rowClick", self._row_click)

    def update(self):
        """Update the logs table and Process input list"""
        self._update_table()
        self._update_process_input()

    def _update_table(self):
        """Update the table with logs from the database applying the filters."""
        from_date = self.from_input.get_datetime()
        to_date = self.to_input.get_datetime()
        process_name = self.process_input.value if self.process_input.value != 'All' else None
        level = LogLevel(self.level_input.value) if self.level_input.value != "All" else None
        limit = self.limit_input.value

        logs = db_util.get_logs(0, limit=limit, from_date=from_date, to_date=to_date, log_level=level, process_name=process_name)
        self.logs_table.rows = [log.to_row_dict() for log in logs]

    def _update_process_input(self):
        """Update the process input with names from the database."""
        process_names = list(db_util.get_unique_log_process_names())
        process_names.insert(0, "All")
        self.process_input.options = process_names
        self.process_input.update()

    def _row_click(self, event):
        """Display a dialog with info on the clicked log."""
        row = event.args[1]
        with ui.dialog(value=True), ui.card():
            ui.label("Log ID:").classes("font-bold")
            ui.label(row['ID'])
            ui.label("Log Time:").classes("font-bold")
            ui.label(row['Log Time'])
            ui.label("Process Name:").classes("font-bold")
            ui.label(row['Process Name'])
            ui.label("Log Level:").classes("font-bold")
            ui.label(row['Level'])
            ui.label("Message:").classes("font-bold")
            ui.html(f"<pre>{row['Message']}</pre>")



"""This module is used to run Orchestrator or Scheduler from the command line.
Usage: python -m OpenOrchestrator [-o|-s]"""

import sys
from OpenOrchestrator.scheduler.application import Application as s_app
from OpenOrchestrator.orchestrator.application import Application as o_app


def _print_usage():
    print("Usage: -o to start Orchestrator. -s to start Scheduler")


if len(sys.argv) != 2:
    _print_usage()
elif sys.argv[1] == '-s':
    s_app()
elif sys.argv[1] == '-o':
    o_app()
else:
    _print_usage()



"""This module defines ORM classes for constants and credentials."""

from datetime import datetime

from sqlalchemy import String, Engine
from sqlalchemy.orm import Mapped, DeclarativeBase, mapped_column

from OpenOrchestrator.common import datetime_util

# All classes in this module are effectively dataclasses without methods.
# pylint: disable=too-few-public-methods


class Base(DeclarativeBase):
    """SqlAlchemy base class for all ORM classes in this module."""


class Constant(Base):
    """A class representing a constant object in the ORM."""
    __tablename__ = "Constants"

    name: Mapped[str] = mapped_column(String(100), primary_key=True)
    value: Mapped[str] = mapped_column(String(1000))
    changed_at: Mapped[datetime] = mapped_column(onupdate=datetime.now, default=datetime.now)

    def to_row_dict(self) -> dict[str, str]:
        """Convert constant to a row dictionary for display in a table."""
        return {
            "Constant Name": self.name,
            "Value": self.value,
            "Last Changed": datetime_util.format_datetime(self.changed_at)
        }


class Credential(Base):
    """A class representing a credential object in the ORM."""
    __tablename__ = "Credentials"

    name: Mapped[str] = mapped_column(String(100), primary_key=True)
    username: Mapped[str] = mapped_column(String(250))
    password: Mapped[str] = mapped_column(String(1000))
    changed_at: Mapped[datetime] = mapped_column(onupdate=datetime.now, default=datetime.now)

    def to_row_dict(self) -> dict[str, str]:
        """Convert credential to a row dictionary for display in a table."""
        return {
            "Credential Name": self.name,
            "Username": self.username,
            "Password": self.format_password(),
            "Last Changed": datetime_util.format_datetime(self.changed_at)
        }

    def format_password(self) -> str:
        """Format the password to be shown in a table."""
        length = len(self.password)
        lower_length = int(((length-100)/20)*16)
        upper_length = lower_length + 15
        return f"{length} encrypted bytes. {lower_length}-{upper_length} decrypted bytes."


def create_tables(engine: Engine):
    """Create all SQL tables related to ORM classes in this module.

    Args:
        engine: The SqlAlchemy connection engine used to create the tables.
    """
    Base.metadata.create_all(engine)



"""This module is responsible for the layout and functionality of the settings tab
in Scheduler."""

from tkinter import ttk

from OpenOrchestrator.scheduler.connection_frame import ConnectionFrame


def create_tab(parent: ttk.Notebook) -> ttk.Frame:
    """Creates a new Settings tab object.

    Args:
        parent (ttk.Notebook): The ttk.Notebook that this tab is a child of.

    Returns:
        ttk.Frame: The created tab object as a ttk.Frame.
    """
    tab = ttk.Frame(parent)
    tab.pack(fill='both', expand=True)

    conn_frame = ConnectionFrame(tab)
    conn_frame.pack(fill='x')

    return tab



"""This module is responsible for the layout and functionality of the Settings tab
in Orchestrator."""

from nicegui import ui

from OpenOrchestrator.database import db_util
from OpenOrchestrator.common.connection_frame import ConnectionFrame


# pylint: disable-next=too-few-public-methods
class SettingsTab():
    """The settings tab object for Orchestrator."""
    def __init__(self, tab_name: str) -> None:
        with ui.tab_panel(tab_name):
            conn_frame = ConnectionFrame()
            with ui.row().classes("w-full"):
                self.key_button = ui.button("Generate Key", on_click=conn_frame.new_key)
                self.init_button = ui.button("Initialize Database", on_click=self._init_database)

    def _init_database(self):
        db_util.initialize_database()
        ui.notify("Database initialized!", type='positive')



"""This module is responsible for checking triggers and running processes."""

import os
import subprocess
from dataclasses import dataclass
import uuid

from OpenOrchestrator.common import crypto_util
from OpenOrchestrator.database import db_util
from OpenOrchestrator.database.triggers import Trigger, SingleTrigger, ScheduledTrigger, QueueTrigger, TriggerStatus
from OpenOrchestrator.database.logs import LogLevel


@dataclass
class Job():
    """An object that holds information about a running job."""
    process: subprocess.Popen
    trigger: Trigger
    process_folder: str | None


def poll_triggers(app) -> Job | None:
    """Checks if any triggers are waiting to run. If any the first will be run and a
    corresponding job object will be returned.

    Args:
        app: The Application object of the Scheduler app.

    Returns:
        Job: A job object describing the job that has been launched, if any else None.
    """

    other_processes_running = len(app.running_jobs) != 0

    # Single triggers
    next_single_trigger = db_util.get_next_single_trigger()

    if next_single_trigger and not (next_single_trigger.is_blocking and other_processes_running):
        return run_single_trigger(next_single_trigger)

    # Scheduled triggers
    next_scheduled_trigger = db_util.get_next_scheduled_trigger()

    if next_scheduled_trigger and not (next_scheduled_trigger.is_blocking and other_processes_running):
        return run_scheduled_trigger(next_scheduled_trigger)

    # Queue triggers
    next_queue_trigger = db_util.get_next_queue_trigger()

    if next_queue_trigger and not (next_queue_trigger.is_blocking and other_processes_running):
        return run_queue_trigger(next_queue_trigger)

    return None


def run_single_trigger(trigger: SingleTrigger) -> Job | None:
    """Mark a single trigger as running in the database
    and start the process.

    Args:
        trigger: The trigger to run.

    Returns:
        Job: A Job object describing the process if successful.
    """

    print('Running trigger: ', trigger.trigger_name)

    if db_util.begin_single_trigger(trigger.id):
        return run_process(trigger)

    return None


def run_scheduled_trigger(trigger: ScheduledTrigger) -> Job | None:
    """Mark a scheduled trigger as running in the database,
    calculate the next run datetime,
    and start the process.

    Args:
        trigger: The trigger to run.

    Returns:
        Job: A Job object describing the process if successful.
    """
    print('Running trigger: ', trigger.trigger_name)

    if db_util.begin_scheduled_trigger(trigger.id):
        return run_process(trigger)

    return None


def run_queue_trigger(trigger: QueueTrigger) -> Job | None:
    """Mark a queue trigger as running in the database
    and start the process.

    Args:
        trigger: The trigger to run.

    Returns:
        Job: A Job object describing the process if successful.
    """
    print('Running trigger: ', trigger.trigger_name)

    if db_util.begin_queue_trigger(trigger.id):
        return run_process(trigger)

    return None


def clone_git_repo(repo_url: str) -> str:
    """Clone the git repo at the path to %USER%\\desktop\\Scheduler_Repos\\%UUID%.

    Args:
        repo_url: URL to the git repo to clone.

    Returns:
        str: The path to the cloned repo on the desktop.
    """
    repo_folder = get_repo_folder_path()
    unique_id = str(uuid.uuid4())
    repo_path = os.path.join(repo_folder, unique_id)

    os.makedirs(repo_path)
    try:
        subprocess.run(['git', 'clone', repo_url, repo_path], check=True)
    except subprocess.CalledProcessError as exc:
        raise ValueError(f"Failed to clone git repo at {repo_url}.") from exc

    return repo_path


def clear_repo_folder() -> None:
    """Completely remove the repos folder."""
    repo_folder = get_repo_folder_path()
    clear_folder(repo_folder)


def get_repo_folder_path() -> str:
    """Gets the path to the folder where robot repos should be saved.

    Returns:
        str: The absolute path to the repo folder.
    """
    user_path = os.path.expanduser("~")
    repo_path = os.path.join(user_path, "Desktop", "Scheduler_Repos")
    return repo_path


def clear_folder(folder_path: str) -> None:
    """Clear a folder on the system.

    Args:
        folder_path: The folder to remove.
    """
    subprocess.run(['rmdir', '/s', '/q', folder_path], check=False, shell=True, capture_output=True)


def find_main_file(folder_path: str) -> str:
    """Finds the file in the given folder with the name 'main.py'.
    The search checks subfolders recursively.
    Only the first found file is returned.

    Args:
        folder_path: The path to the folder.

    Returns:
        str: The path to the main.py file.
    """
    for dir_path, _, file_names in os.walk(folder_path):
        for file_name in file_names:
            if file_name == 'main.py':
                return os.path.join(dir_path, file_name)

    raise ValueError("No 'main.py' file found in the folder or its subfolders.")


def end_job(job: Job) -> None:
    """Mark a job as ended in the triggers table
    in the database.
    If it's a single trigger it's marked as 'Done'
    else it's marked as 'Idle' or 'Paused'.

    Args:
        job: The job whose trigger to mark as ended.
    """
    if isinstance(job.trigger, SingleTrigger):
        db_util.set_trigger_status(job.trigger.id, TriggerStatus.DONE)

    elif isinstance(job.trigger, (ScheduledTrigger, QueueTrigger)):
        current_status = db_util.get_trigger(job.trigger.id).process_status
        if current_status == TriggerStatus.PAUSING:
            db_util.set_trigger_status(job.trigger.id, TriggerStatus.PAUSED)
        elif current_status == TriggerStatus.RUNNING:
            db_util.set_trigger_status(job.trigger.id, TriggerStatus.IDLE)

    if job.process_folder:
        clear_folder(job.process_folder)


def fail_job(job: Job) -> None:
    """Mark a job as failed in the triggers table in the database.

    Args:
        job: The job whose trigger to mark as failed.
    """
    db_util.set_trigger_status(job.trigger.id, TriggerStatus.FAILED)
    _, error = job.process.communicate()
    error_msg = f"An uncaught error ocurred during the process:\n{error}"
    db_util.create_log(job.trigger.process_name, LogLevel.ERROR, error_msg)

    if job.process_folder:
        clear_folder(job.process_folder)


def run_process(trigger: Trigger) -> Job | None:
    """Runs the process of the given trigger with the necessary inputs:
    Process name
    Connection string
    Crypto key
    Process args

    If the trigger's process_path is pointing to a git repo the repo is cloned
    and the main.py file in the repo is found and run.

    Supports only .py files.

    If any exceptions occur during launch the trigger will be marked as failed.

    Args:
        trigger: The trigger whose process to run.

    Returns:
        Job: A Job object referencing the process if succesful.
    """
    process_path = trigger.process_path
    folder_path = None

    try:
        if trigger.is_git_repo:
            folder_path = clone_git_repo(process_path)
            process_path = find_main_file(folder_path)

        if not os.path.isfile(process_path):
            raise ValueError(f"The process path didn't point to a file on the system. Path: '{process_path}'")

        if not process_path.endswith(".py"):
            raise ValueError(f"The process path didn't point to a valid file. Supported files are [.py]. Path: '{process_path}'")

        conn_string = db_util.get_conn_string()
        crypto_key = crypto_util.get_key()

        command_args = ['python', process_path, trigger.process_name, conn_string, crypto_key, trigger.process_args]

        process = subprocess.Popen(command_args, stderr=subprocess.PIPE, text=True)  # pylint: disable=consider-using-with

        return Job(process, trigger, folder_path)

    # We actually want to catch any exception here
    # pylint: disable=broad-exception-caught
    except Exception as exc:
        db_util.set_trigger_status(trigger.id, TriggerStatus.FAILED)
        error_msg = f"Scheduler couldn't launch the process:\n{exc.__class__.__name__}:\n{exc}"
        db_util.create_log(trigger.process_name, LogLevel.ERROR, error_msg)
        print(error_msg)

    return None
